from distutils.core import setup, Extension
from Cython.Distutils import build_ext
from Cython.Build import cythonize

import numpy
import os

try:
    os.remove("ampere/models/P2D/P2D_fd.c")
except FileNotFoundError:
    pass

# ida_dir = "../ida"
# setup(
ida_dir = "ampere/models/ida"
#     cmdclass={'build_ext': build_ext},
#     ext_modules=[Extension("P2D_fd", ["P2D_fd.pyx", ida_dir+"/ida.c",ida_dir+"/ida_band.c",ida_dir+"/ida_dense.c",ida_dir+"/ida_direct.c",ida_dir+"/ida_ic.c",ida_dir+"/ida_io.c",
# 				ida_dir+"/nvector_serial.c",ida_dir+"/sundials_band.c",ida_dir+"/sundials_dense.c",ida_dir+"/sundials_direct.c",ida_dir+"/sundials_math.c",ida_dir+"/sundials_nvector.c"],
# 							include_dirs=[numpy.get_include(), ida_dir])]
# )


#
extension = Extension(
    name="ampere.models.P2D.P2D_fd",
    sources=["ampere/models/P2D/P2D_fd.pyx", ida_dir+"/ida.c",ida_dir+"/ida_band.c",ida_dir+"/ida_dense.c",ida_dir+"/ida_direct.c",ida_dir+"/ida_ic.c",ida_dir+"/ida_io.c",
    				ida_dir+"/nvector_serial.c",ida_dir+"/sundials_band.c",ida_dir+"/sundials_dense.c",ida_dir+"/sundials_direct.c",ida_dir+"/sundials_math.c",ida_dir+"/sundials_nvector.c"],
    library_dirs=['ida'],
    include_dirs=[numpy.get_include(), ida_dir])

setup(
    name='ampere.models.P2D.P2D_fd',
    # cmdclass={'build_ext': build_ext},
    ext_modules=cythonize([extension])
)



import numpy as np
from typing import Tuple, List
from .SPM_par import model as SPM_par
from .SPM_fd_sei import model as SPM_fd_sei
from .SPM_fd import model as SPM_fd


def spm_parabolic(p: np.ndarray, initial_state=(), tf=0, internal: bool = False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    # initial_state is ignored if it is not defined
    model_inputs = np.concatenate([p, [0], [tf], initial_state])
    result_dimensionality = len(initial_state) + 1
    expected_max_result_length = 10000

    # pre-allocate the array for the C library to fill
    model_results = np.zeros((expected_max_result_length, result_dimensionality))
    SPM_par(model_inputs, model_results)

    # throw out the extra values
    count = np.nonzero(model_results[:, 1])[0][-1] + 1
    model_results = model_results[:count]

    final_state = model_results[-1]
    time_ind, pos_potential_ind, neg_potential_ind, curr_ind = 0, 5, 6, 7
    out = model_results[:, [time_ind, pos_potential_ind, curr_ind]]

    # calculate the external voltage from the internal potential differences
    out[:, 1] -= model_results[:, neg_potential_ind]
    return out, final_state, None


def spm_fd_sei(p, initial_state=None, tf=0, internal=False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    model_inputs = np.concatenate([p, [0], [tf], initial_state])
    result_dimensionality = len(initial_state) + 1
    # print(len(initial_state), result_dimensionality)
    var = np.zeros((10000, result_dimensionality))
    SPM_fd_sei(model_inputs, var)
    count = np.nonzero(var[:, -2])[0][-1] + 1
    var = var[:count]
    final = var[-1]
    # need to select: time, voltage, current
    out = var[:, [0, -2, -1]]
    out[:, -1] /= 30.0
    # out[:,1] -= var[:,6]
    return out, final, var


def spm_fd(p, initial_state=None, tf=0, internal=False) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    model_inputs = np.concatenate([p, [0], [tf], initial_state])
    result_dimensionality = int(p[14] + p[15] + 9)

    var = np.zeros((10000, result_dimensionality))
    # var = np.zeros((10000, ))
    SPM_fd(model_inputs, var)
    count = np.nonzero(var[:, -2])[0][-1] + 1
    var = var[:count]
    final = var[-1]
    # need to select: time, voltage, current
    out = var[:, [0, -2, -1]]
    out[:, -1] /= 30.0

    return out, final, var



import numpy as np

from typing import List, Dict

from .base_battery import BaseBattery, SimulationResult, ChargeResult


class SingleParticleParabolic(BaseBattery):
    """An implementation of the Single Particle Model, solved with IDA.  This
    version supports CC-discharge, CC-CV charging, and access to the internal
    states of the battery.  It also allows for sequential cycling."""

    def __init__(self, initial_parameters: Dict[str, float] = None, estimate_parameters : List[str] = None, verbose=False): # noqa
        """Constructor for the Single Particle Model class with the parabolic intra-particle concentration approximation.

        Parameters
        ----------
        initial_parameters: A dictionary of parameter names and values. Acceptable names for the
            parameters can be found below:
            | name   | description                                 | default value | Units           |
            |--------|---------------------------------------------|---------------|-----------------|
            | Dn     | Li+ Diffusivity in negative particle        | 3.9e-14       | cm^2/s          |
            | Dp     | Li+ Diffusivity in positive particle        | 1e-14         | cm^2/s          |
            | Rn     | Negative particle radius                    | 2e-6          | m               |
            | Rp     | Positive particle radius                    | 2e-6          | m               |
            | T      | Ambient Temperature                         | 303.15        | K               |
            | an     | Surface area of negative electrode          | 723600        | m^2/m^3         |
            | ap     | Surface area of positive electrode          | 885000        | m^2/m^3         |
            | ce     | Starting electrolyte Li+ concentration      | 1000          | mol/m^3         |
            | csnmax | Maximum Li+ concentration of negative solid | 30555         | mol/m^3         |
            | cspmax | Maximum Li+ concentration of positive solid | 51555         | mol/m^3         |
            | kn     | Negative electrode reaction rate            | 5.0307e-9     |m^2.5/(mol^0.5s) |
            | kp     | Positive electrode reaction rate            | 2.334e-9      |m^2.5/(mol^0.5s) |
            | ln     | Negative electrode thickness                | 88e-6         | m               |
            | lp     | Positive electrode thickness                | 80e-6         | m               |

        estimate_parameters: List[str]
            A list of parameters that this instance is allowed to estimate while fitting. Having each instance
                of a model only allow certain values to be estimated helps encourage good code organization.

        Example usage:
        spm = SingleParticleParabolic(initial_parameters={'Rp': 1e-5})
        trial1 = spm.charge(current=1.0)
        trial2 = spm.discharge(current=1.0)
        """
        from .numerical import spm_parabolic
        super().__init__(verbose)
        self.model = spm_parabolic
        self.initial_parameters = {
            "Dn": 3.9e-14,
            "Dp": 1e-14,
            "Rn": 2e-6,
            "Rp": 2e-6,
            "T": 303.15,
            "an": 723600,
            "ap": 885000,
            "ce": 1000,
            "csnmax": 30550,
            "cspmax": 51555,
            "kn": 5.0307e-9,
            "kp": 2.334e-9,
            "ln": 88e-6,
            "lp": 80e-6,
        }
        self.available_parameters = list(self.initial_parameters.keys())

        if isinstance(initial_parameters, dict):
            for key in initial_parameters.keys():
                assert key in self.initial_parameters, "Invalid initial key entered - double check %s" % str(key)
                self.initial_parameters[key] = initial_parameters[key]

        # initial enforces the order parameters are given to the model
        # as of python 3.6, the order of keys in a dictionary can be relied upon due to an implementation detail
        self.initial = np.array([self.initial_parameters[i] for i in self.available_parameters])
        self._validate_estimate_parameters(estimate_parameters)

        # there is no internal structure - the parabolic assumption of internal concentration
        # gradients prevents peering into the particles.
        self.internal_structure = {}

        self.charge_ICs = [
            4.95030611e04,
            3.05605527e02,
            4.93273985e04,
            3.55685791e02,
            3.78436346e00,
            7.86330739e-01,
            1.00000000e00,
        ]
        self.discharge_ICs = [
            2.51584754e04,
            2.73734963e04,
            2.51409091e04,
            2.73785043e04,
            4.26705391e00,
            6.70539113e-02,
            -1.00000000,
        ]


class SingleParticleFDSEI(BaseBattery):
    """An Finite Difference implementation of the Single Particle Model with SEI, solved with IDA.  This
    version supports CC-discharge, CC-CV charging, and access to the internal
    states of the battery.  It also allows for sequential cycling. See .charge, .discharge,
    .cycle, and .piecewise_current for more."""

    def __init__(self, initial_parameters: Dict[str, float]=None, estimate_parameters: List[str] = None, verbose=False):
        """Constructor for the Single Particle Model class with Finite Difference discretization and SEI layer building
        Parameters
        ----------
        initial_parameters: Dict[str, float]
            A dictionary of parameter names and values. Acceptable names for the parameters can be found below:
            | name      | description                                   | default value | Units            |
            |-----------|-----------------------------------------------|---------------|------------------|
            | Dp        | Li+ Diffusivity in positive particle          | 1e-14        | cm^2/s           |
            | Dn        | Li+ Diffusivity in negative particle          | 1e-14        | cm^2/s           |
            | cspmax    | Maximum Li concentration of positive solid    | 51555         | mol/m^3          |
            | csnmax    | Maximum Li concentration of negative solid    | 30555         | mol/m^3          |
            | lp        | Positive electrode thickness                  | 80e-6         | m                |
            | ln        | Negative electrode thickness                  | 88e-6         | m                |
            | Rp        | Positive particle radius                      | 2e-6          | m                |
            | Rn        | Negative particle radius                      | 2e-6          | m                |
            | T         | Ambient Temperature                           | 303.15        | K                |
            | ce        | Starting electrolyte Li+ concentration        | 1000          | mol/m^3          |
            | ap        | Surface area of positive electrode per volume | 885000        | m^2/m^3          |
            | an        | Surface area of negative electrode per volume | 723600        | m^2/m^3          |
            | M_sei     | Molecular weight of SEI                       | 0.026         | Kg/mol           |
            | rho_sei   | SEI Density                                   | 2.1e3         | Kg/m^3           |
            | Kappa_sei | SEI Ionic conductivity                        | 1             | S/m              |
            | k_sei     | rate constant of side reaction                | 1.5e-6        | C m/(mol*s)      |
            | kp        | Positive electrode reaction rate              | 2.334e-9      | m^2.5/(mol^0.5s) |
            | kn        | Negative electrode reaction rate              | 5.0307e-9     | m^2.5/(mol^0.5s) |
            | N1        | Number of FD nodes in positive particle       | 30            |                  |
            | N2        | Number of FD nodes in negative particle       | 30            |                  |

        estimate_parameters: List[str]
            A list of strings representing the parameters that you wish to estimate.
            Defaults to None, which will allow for the estimation of all parameters except temperature.
            For both initial_parameters and estimate_parameters, order does not matter.

        Example usage:
        Example usage:
        spm = SingleParticleFDSEI(initial_parameters={'Rp': 1e-5})
        trial1 = spm.charge(current=1.0)
        trial2 = spm.discharge(current=1.0)
        """
        from .numerical import spm_fd_sei
        super().__init__(verbose)
        self.model = spm_fd_sei
        self.opt = self.opt_wrap
        self.verbose = verbose
        self.initial_fit = 0

        self.initial_parameters = {
            "Dp": 1e-14,
            "Dn": 1e-14,
            "cspmax": 51555,
            "csnmax": 30555,
            "lp": 80e-6,
            "ln": 88e-6,
            "Rp": 2e-6,
            "Rn": 2e-6,
            "T": 303.15,
            "ce": 1000,
            "ap": 885000,
            "an": 723600,
            "M_sei": 0.026,
            "rho_sei": 2100,
            "Kappa_sei": 1.0,
            "kp": 2.334e-11,
            "kn": 8.307e-12,
            "ksei": 1.5e-06,
            "N1": 30,
            "N2": 30,
        }

        self.available_parameters = list(self.initial_parameters.keys())
        if isinstance(initial_parameters, dict):
            for key in initial_parameters.keys():
                assert key in self.initial_parameters, "Invalid initial key entered - double check %s" % str(key)
                self.initial_parameters[key] = initial_parameters[key]

        # initial enforces the order parameters are given to the model
        # as of python 3.6, the order of keys in a dictionary can be relied upon due to an implementation detail
        self.initial = np.array([self.initial_parameters[i] for i in self.available_parameters])
        self._validate_estimate_parameters(estimate_parameters)

        n_positive = self.initial_parameters["N1"]
        n_negative = self.initial_parameters["N2"]

        self.internal_structure = {
            'time': [0],
            'positive_concentration': list(range(1, n_positive + 3)),
            'negative_concentration': list(range(n_positive + 3, n_positive + 3 + n_negative + 2)),
            'SEI thickness': [3 + n_positive + n_negative + 2 + 5]
        }

        # set initial lithium concentrations inside the particles
        self.charge_ICs = [49503.111 for n in range(n_positive + 2)] + [305.55 for n in range(n_negative + 2)]
        self.charge_ICs += [
            3.67873289259766,   # phi_p
            0.182763748093840,  # phi_n
            30,                 # iint
            0,                  # isei
            1e-10,              # delta_sei
            0,                  # Q
            0,                  # cm
            0,                  # cf
            3.0596914450382,    # potential
            30,                 # it
        ]

        # set initial lithium concentrations inside the particles
        self.discharge_ICs = [2.51417672e04 for n in range(n_positive + 2)] + [2.73921225e04 for n in range(n_negative + 2)]
        self.discharge_ICs += [
            4.26700382e00,      # phi_p
            6.70038247e-02,     # phi_n
            2.65295200e-03,     # iint,
            7.34704800e-03,     # isei
            1e-10,              # delta_sei
            3.08271510e01,      # Q
            3.08183958e01,      # cm
            8.75512593e-03,     # cf
            4.20000000e00,      # pot
            30,                 # it
        ]


class SingleParticleFD(BaseBattery):
    """An Finite Difference implementation of the Single Particle Model with SEI, solved with IDA.  This
    version supports CC-discharge, CC-CV charging, and access to the internal
    states of the battery.  It also allows for sequential cycling. See .charge, .discharge,
    .cycle, and .piecewise_current for more."""

    def __init__(self, initial_parameters: Dict[str, float] = None, estimate_parameters: List[str] = None, verbose=False):
        """Constructor for the Single Particle Model class with Finite Difference discretization.
        Parameters
        ----------
        initial_parameters: Dict[str, float]
            A dictionary of parameter names and values. Acceptable names for the parameters can be found below:
            | name      | description                                   | default value | Units            |
            |-----------|-----------------------------------------------|---------------|------------------|
            | Dp        | Li+ Diffusivity in positive particle          | 1e-14        | cm^2/s           |
            | Dn        | Li+ Diffusivity in negative particle          | 1e-14        | cm^2/s           |
            | cspmax    | Maximum Li concentration of positive solid    | 51555         | mol/m^3          |
            | csnmax    | Maximum Li concentration of negative solid    | 30555         | mol/m^3          |
            | lp        | Positive electrode thickness                  | 80e-6         | m                |
            | ln        | Negative electrode thickness                  | 88e-6         | m                |
            | Rp        | Positive particle radius                      | 2e-6          | m                |
            | Rn        | Negative particle radius                      | 2e-6          | m                |
            | T         | Ambient Temperature                           | 303.15        | K                |
            | ce        | Starting electrolyte Li+ concentration        | 1000          | mol/m^3          |
            | ap        | Surface area of positive electrode per volume | 885000        | m^2/m^3          |
            | an        | Surface area of negative electrode per volume | 723600        | m^2/m^3          |
            | M_sei     | Molecular weight of SEI                       | 0.026         | Kg/mol           |
            | rho_sei   | SEI Density                                   | 2.1e3         | Kg/m^3           |
            | Kappa_sei | SEI Ionic conductivity                        | 1             | S/m              |
            | k_sei     | rate constant of side reaction                | 1.5e-6        | C m/(mol*s)      |
            | kp        | Positive electrode reaction rate              | 2.334e-9      | m^2.5/(mol^0.5s) |
            | kn        | Negative electrode reaction rate              | 5.0307e-9     | m^2.5/(mol^0.5s) |
            | N1        | Number of FD nodes in positive particle       | 30            |                  |
            | N2        | Number of FD nodes in negative particle       | 30            |                  |

        estimate_parameters: List[str]
            A list of strings representing the parameters that you wish to estimate.
            Defaults to None, which will allow for the estimation of all parameters except temperature.
            For both initial_parameters and estimate_parameters, order does not matter.

        Example usage:
        Example usage:
        spm = SingleParticleFDSEI(initial_parameters={'Rp': 1e-5})
        trial1 = spm.charge(current=1.0)
        trial2 = spm.discharge(current=1.0)
        """
        from .numerical import spm_fd
        super().__init__(verbose)
        self.model = spm_fd
        self.verbose = verbose
        self.initial_fit = 0

        self.initial_parameters = {
            "Dp": 1e-14,
            "Dn": 1e-14,
            "cspmax": 51555,
            "csnmax": 30555,
            "lp": 80e-6,
            "ln": 88e-6,
            "Rp": 2e-6,
            "Rn": 2e-6,
            "T": 303.15,
            "ce": 1000,
            "ap": 885000,
            "an": 723600,
            "kp": 2.334e-11,
            "kn": 8.307e-12,
            "N1": 30,
            "N2": 30,
        }

        self.available_parameters = list(self.initial_parameters.keys())
        if isinstance(initial_parameters, dict):
            for key in initial_parameters.keys():
                assert key in self.initial_parameters, "Invalid initial key entered - double check %s" % str(key)
                self.initial_parameters[key] = initial_parameters[key]

        # initial enforces the order parameters are given to the model
        # as of python 3.6, the order of keys in a dictionary can be relied upon due to an implementation detail
        self.initial = np.array([self.initial_parameters[i] for i in self.available_parameters])
        self._validate_estimate_parameters(estimate_parameters)

        n_positive = self.initial_parameters["N1"]
        n_negative = self.initial_parameters["N2"]

        self.internal_structure = {
            'time': [0],
            'positive_concentration': list(range(1, n_positive + 3)),
            'negative_concentration': list(range(n_positive + 3, n_positive + 3 + n_negative + 2))
        }

        # set initial lithium concentrations
        self.charge_ICs = [49503.111 for n in range(n_positive + 2)] + [305.55 for n in range(n_negative + 2)]
        self.charge_ICs += [
            3.67873289259766,  # phi_p
            0.182763748093840,  # phi_n
            3.0596914450382,  # pot
            30,  # it
        ]

        self.discharge_ICs = [25817.37 for n in range(n_positive + 2)] + [26885.03 for n in range(n_negative + 2)]
        self.discharge_ICs += [
            4.246347,  # phi_p
            0.046347,  # phi_n
            4.20000000e00,  # pot
            30,  # it
        ]


class PseudoTwoDimFD(BaseBattery):
    """An Finite Difference implementation of the Single Particle Model with SEI, solved with IDA.  This
    version supports CC-discharge, CC-CV charging, and access to the internal
    states of the battery.  It also allows for sequential cycling. See .charge, .discharge,
    .cycle, and .piecewise_current for more."""

    def __init__(self, initial_parameters: Dict[str, float] = None, estimate_parameters: List[str] = None, verbose=False):
        """Constructor for the Single Particle Model base class
        Parameters
        ----------
        initial_parameters: A dictionary of parameter names and values. Acceptable names for the
        parameters can be found below:
        |     name    |                           description                          |  default value  | Units              |
        | ----------- | -----------------------------------------------                | --------------- | ------------------ |
        | D1          | Li+ Diffusivity in electrolyte                                 | 0.15e-8         | cm^2/s             |
        | Dp          | Li+ Diffusivity in positive particle                           | 7.2e-14         | cm^2/s             |
        | Dn          | Li+ Diffusivity in negative particle                           | 7.5e-14         | cm^2/s             |
        | cspmax      | Maximum Li concentration of positive solid                     | 45829           | mol/m^3            |
        | csnmax      | Maximum Li concentration of negative solid                     | 30555           | mol/m^3            |
        | ls          | Separator thickness                                            | 16e-6           | m                  |
        | lp          | Positive electrode thickness                                   | 43e-6           | m                  |
        | ln          | Negative electrode thickness                                   | 46.5e-6         | m                  |
        | es          | Separator volume fraction of pores                             | 0.45            | m^3/m^3            |
        | ep          | Positive electrode volume fraction of pores                    | 0.4             | m^3/m^3            |
        | en          | Negative electrode volume fraction of pores                    | 0.38            | m^3/m^3            |
        | efn         | Negative electrode filler fraction                             | 0.0326          | m^3/m^3            |
        | efp         | Positive electrode filler fraction                             | 0.025           | m^3/m^3            |
        | brugs       | Separator Bruggeman coefficient - pore tortuosity              | 1.5             |                    |
        | brugn       | Negative electrode Bruggeman coefficient - pore tortuosity     | 1.5             |                    |
        | brugp       | Positive electrode coefficient - pore tortuosity               | 1.5             |                    |
        | sigma_n     | Negative electrode electrical conductivity                     | 100             | S/m                |
        | sigma_p     | Positive electrode electrical conductivity                     | 10              | S/m                |
        | t+          | Transference Number - fraction of ionic current carried by Li+ | 0.363           |                    |
        | Rp          | Positive particle radius                                       | 10e-6           | m                  |
        | Rn          | Negative particle radius                                       | 8e-6            | m                  |
        | T           | Ambient Temperature                                            | 303.15          | K                  |
        | ce          | Starting electrolyte Li+ concentration                         | 1200            | mol/m^3            |
        | ap          | Surface area of positive electrode per volume                  | 885000          | m^2/m^3            |
        | an          | Surface area of negative electrode per volume                  | 723600          | m^2/m^3            |
        | kp          | Positive electrode reaction rate                               | 0.10307e-9      | m^2.5/(mol^0.5s)   |
        | kn          | Negative electrode reaction rate                               | 0.1334e-9       | m^2.5/(mol^0.5s)   |
        | N1          | Positive electrode number of FD Nodes                          | 7               |                    |
        | N2          | Separator number of FD Nodes                                   | 3               |                    |
        | N3          | Negative electrode number of FD Nodes                          | 7               |                    |
        | Nr1         | Positive particle number of FD nodes (per particle)            | 3               |                    |
        | Nr2         | Negative particle number of FD nodes (per particle)            | 3               |                    |

        estimate_parameters: A list of strings representing the parameters that you wish to estimate.
        Defaults to None, which will allow for the estimation of all parameters except temperature.

        For both intiial_parameters and estimate_parameters, order does not matter.

        Example usage:
        spm = SingleParticle(initial_parameters=dictionary_of_parameter_label_value_pairs, est_pars=list_of_parameter_labels)

        A list of available keyword agruments (kwargs):



        """
        from .numerical import p2d_fd
        super().__init__(initial_parameters)
        self.model = p2d_fd
        self.verbose = verbose
        self.initial_fit = 0

        self.initial_parameters = {
            "D1": 0.15e-8,
            "Dsn": 7.2e-14,
            "Dsp": 7.5e-14,
            "Rn": 10e-6,
            "Rp": 8e-6,
            "T": 303.15,
            "brugn": 1.5,
            "brugp": 1.5,
            "brugs": 1.5,
            "c0": 1200,
            "csnmax": 30555,
            "cspmax": 45829,
            "efn": 0.0326,
            "efp": 0.025,
            "en": 0.38,
            "ep": 0.4,
            "es": 0.45,
            "kn": 0.10307e-9,
            "kp": 0.1334e-9,
            "ln": 46.5e-6,
            "lp": 43e-6,
            "ls": 16e-6,
            "sigma_n": 100,
            "sigma_p": 10,
            "t1": 0.363,
            "N1": 7,
            "N2": 3,
            "N3": 7,
            "Nr1": 3,
            "Nr2": 3,
        }
        self.available_parameters = list(self.initial_parameters.keys())
        if isinstance(initial_parameters, dict):
            for key in initial_parameters.keys():
                assert key in self.initial_parameters, "Invalid initial key entered - double check %s" % str(key)
                self.initial_parameters[key] = initial_parameters[key]

        # initial enforces the order parameters are given to the model
        # as of python 3.6, the order of keys in a dictionary can be relied upon due to an implementation detail
        self.initial = np.array([self.initial_parameters[i] for i in self.available_parameters])
        self._validate_estimate_parameters(estimate_parameters)

        n_positive_region = int(self.initial[25])
        n_separator_region = int(self.initial[26])
        n_negative_region = int(self.initial[27])
        nr_positive = int(self.initial[28])
        nr_negative = int(self.initial[29])

        # here, the internal structure is more complex.
        # for each "point" along the material regions, there is a simulated particle.
        self.internal_structure = {'time': [0], 'potential': [-2], 'current': [-1]}

        # initialize lithium concentration across the electrolyte
        self.internal_structure.update({
            'electrolyte_lithium_concentration': {
                'positive': [n + 2 for n in range(n_positive_region + 1)],
                'separator': [n + 3 + n_positive_region for n in range(n_separator_region + 1)],
                'negative': [n + 4 + n_positive_region + n_separator_region for n in range(n_negative_region + 1)]
            }
        })
        self.charge_ICs = [1.0 for i in range(n_positive_region + n_separator_region + n_negative_region + 4)]

        # initialize liquid-phase potential across the electrolyte. The slight applied gradient helps the model initialize.
        self.internal_structure.update({
            'liquid_phase_potential': {
                'positive': [n + len(self.charge_ICs) + 1 for n in range(n_positive_region + 1)],
                'separator': [n + len(self.charge_ICs) + 2 + n_positive_region for n in range(n_separator_region + 1)],
                'negative': [n + len(self.charge_ICs) + 3 + n_positive_region + n_separator_region for n in range(n_negative_region + 2)]
            }
        })
        self.charge_ICs += [-0.787e-2 + 0.03e-2 * i for i in range(n_positive_region + n_separator_region + n_negative_region + 3)]

        # set liquid-phase potential to be 0 at the negative electrode current collector - this acts as absolute ground
        self.charge_ICs += [0.0]
        # set initial solid-phase potential for each electrode
        self.internal_structure.update({
            'solid_phase_potential': {
                'positive': [n + len(self.charge_ICs) + 1 for n in range(n_positive_region + 2)],
                'negative': [n + len(self.charge_ICs) + 3 + n_positive_region for n in range(n_negative_region + 2)]
            }
        })
        self.charge_ICs += [2.899 for i in range(n_positive_region + 2)]
        self.charge_ICs += [0.09902 for i in range(n_negative_region + 2)]

        # set initial lithium concentration inside the negative electrode
        # in the equations, the pattern is as follows:
        # * the centers of all of the particles are set
        # * the internals of all of the particles are set, grouped by location along the electrode
        # * the externals of all of the particles are set.
        self.internal_structure.update({
            'solid_lithium_concentration': {f'negative_{n}': [n + len(self.charge_ICs) + 1] for n in range(n_negative_region)}
        })
        count = len(self.charge_ICs) + n_negative_region + 1
        for p in range(nr_negative + 1):
            for n in range(n_negative_region):
                self.internal_structure['solid_lithium_concentration'][f'negative_{n}'] += [count]
                count += 1
        self.charge_ICs += [0.22800075826244027 for i in range(n_negative_region * (nr_negative + 1))]
        self.charge_ICs += [0.21325933957011173 for i in range(n_negative_region)]

        self.internal_structure['solid_lithium_concentration'].update({f'positive_{n}': [n + len(self.charge_ICs) + 1] for n in range(n_positive_region)})
        count = len(self.charge_ICs) + n_positive_region + 1
        for p in range(nr_positive + 1):
            for n in range(n_positive_region):
                self.internal_structure['solid_lithium_concentration'][f'positive_{n}'] += [count]
                count += 1

        self.charge_ICs += [0.9532086891149233 for i in range(n_positive_region * (nr_positive + 1))]
        self.charge_ICs += [0.9780166774057617 for i in range(n_positive_region)]  # set the surface to initially lower concentration to help with initialization
        self.charge_ICs.append(2.8)  # potential
        self.charge_ICs.append(-17.1)  # initial discharge current

        # initialize the discharge ICs
        self.discharge_ICs = [0.99998 + 2e-6 * i for i in range(n_positive_region + n_separator_region + n_negative_region + 4)]
        self.discharge_ICs += [-0.3447e-2 + 0.01e-2 * i for i in range(n_positive_region + n_separator_region + n_negative_region + 3)]
        self.discharge_ICs += [0.0]
        self.discharge_ICs += [0.422461225901562e1 for i in range(n_positive_region + 2)]
        self.discharge_ICs += [0.822991162960124e-1 for i in range(n_negative_region + 2)]
        self.discharge_ICs += [0.986699999999968 for i in range(n_negative_region * (nr_negative + 1))]
        self.discharge_ICs += [0.977101677061948 for i in range(n_negative_region)]
        self.discharge_ICs += [0.424 for i in range(n_positive_region * (nr_positive + 1))]
        self.discharge_ICs += [0.431 for i in range(n_positive_region)]
        self.discharge_ICs.append(4.2)  # potential
        self.discharge_ICs.append(-17.1)  # initial current



import numpy as np
import time

from collections import namedtuple
from typing import List, Dict

from .numerical import SimulationResult

ChargeResult = namedtuple('ChargeResult', ('time', 'voltage', 'current', 'internal'))
SampleResult = namedtuple('SampleResult', ('inputs', 'voltages', 'failed', 'currents'))


class BaseBattery:
    """Base class for physics-based battery models"""

    def __init__(
        self,
        verbose: bool
    ):

        self.estimate_parameters = []
        self.initial_parameters = dict()
        self.available_parameters = []
        self.initial_fit = 0
        self.bounds = None
        self.verbose = verbose
        self.initial = None
        self.conf_ = None
        self.inplace = None
        self.hist = []
        self.current_state = np.array([])
        self.charge_ICs = np.array([])
        self.discharge_ICs = np.array([])
        self.estimate_indices = []
        self.model = lambda p, t, initial=None, internal=False: SimulationResult
        self.internal_structure = dict()

        # parameters for fitting
        self.fitted_parameters = []
        self.initial_fit = False
        self.opt_result = None
        self.currents = []
        self.currents_type = []
        self.tol = 0
        self.maxiter = 0
        self.count = 0
        self.parameters_ = None
        self.loglist = []

        self.t_exp = []
        self.v_exp = []

    def _validate_estimate_parameters(self, estimate_parameters: List[str]):
        if estimate_parameters is not None:
            for key in estimate_parameters:
                assert key in self.initial_parameters, f"Invalid estimate key entered - double check {key}, not in {self.initial_parameters.keys()}"
            self.estimate_parameters = estimate_parameters
            self.estimate_indices = [i for i, p in enumerate(self.available_parameters) if p in estimate_parameters]
            message = 'Estimating the following parameters:\n' + '\n'.join(p for p in estimate_parameters)
            print(message)
        else:
            self.estimate_indices = list(range(len(self.initial)))
            if self.verbose:
                message = 'Estimating the following parameters:\n' + '\n'.join(p for p in self.initial_parameters)
                print(message)

    def charge(
        self,
        t=None,
        current=0.5,
        from_current_state=False,
        p=None,
        trim=False,
        internal=False,
    ) -> ChargeResult:
        """The base wrapper for the model, used for simple charging. Automatically
        transitions from cc to cv depending upon final time.

        Inputs              Description
        ------              ------------------------
        t                   A list or numpy array of times to sample at (optional)
        current             A value for charge current
        from_current_state  Whether or not the simulation should start from the
                                current state or not - defaults to False, which
                                simulates the battery from a discharged state
        p                   Defaults to none - present only for wrapping in an optimizer
        trim                Defaults to False, a value of true will remove the padded numbers
                                at the end, where False will allow padded values.

        Output              Description
        ------              ------------------------
        out                 A list of values for [time, voltage, current] of the simulation"""
        if t is None:
            t = np.linspace(0, 5000 * (1 / current), 500)
        if p is None:
            if self.initial_fit == 0:
                p = self.initial
            else:
                p = self.initial
                p[self.estimate_indices] = self.fitted_parameters
        initial_state = self.charge_ICs
        if from_current_state:
            initial_state = self.current_state

        result = self.model([*p, current, 1], t, initial=initial_state, internal=internal)
        self.current_state = result.final_state[1:]
        self.hist.append(result.voltage_data)
        if trim:
            result = SimulationResult(
                self._charge_trim(result.voltage_data),
                result.final_state,
                result.internal_states,
            )

        return ChargeResult(result.voltage_data[0], result.voltage_data[1], result.voltage_data[2], result.internal_states if internal else None)

    def _charge_trim(self, voltage_data):
        return voltage_data[:, voltage_data[2] > 0.01]

    def _discharge_trim(self, voltage_data):
        return voltage_data[:, ~(voltage_data[2] == 0)]

    def discharge(
        self,
        t=None,
        current=0.5,
        from_current_state=False,
        p=None,
        trim=False,
        internal=False,
    ):
        """The base wrapper for the model, used for simple discharging.

        Inputs              Description
        ------              ------------------------
        t                   A list or numpy array of times to sample at (optional)
        current             A value for charge current
        from_current_state  Whether or not the simulation should start from the
                                current state or not - defaults to False, which
                                simulates the battery from a discharged state
        p                   Defaults to none - present only for wrapping in an optimizer
        trim                Defaults to False, a value of true will remove the padded numbers
                                at the end, where False will allow padded values.

        Output              Description
        ------              ------------------------
        out                 A list of values for [time, voltage, current] of the simulation"""
        if t is None:
            t = np.linspace(0, 4000 * (1 / current), 500)
        if p is None:
            if self.initial_fit == 0:
                p = self.initial
            else:
                p = self.initial
                p[self.estimate_indices] = self.fitted_parameters
        initial_state = self.discharge_ICs
        if from_current_state:
            initial_state = self.current_state

        result = self.model([*p, current * -1, 1], t, initial=initial_state, internal=internal)
        self.current_state = result.final_state[1:]
        self.hist.append(result.voltage_data)
        if trim:
            result = SimulationResult(
                self._discharge_trim(result.voltage_data),
                result.final_state,
                result.internal_states,
            )
        return ChargeResult(result.voltage_data[0], result.voltage_data[1], result.voltage_data[2], result.internal_states if internal else None)

    def cycle(self, current=0.5, n=500, charge_first=False, p=None, trim=False):
        """This function calls either a charge then discharge, or a discharge followed
        by a change. When charge_first is set to False, it will start from a charged
        state and discharge, follower by a charge.  Otherwise, it will do them
        in reverse order.

        Inputs              Description
        ------              ------------------------
        current             A value for charge and discharge current. These must be the same.
        charge_first        Whether the charge simulation should be run first. Defaults to False.
        n                   The number of points to sample in each charge / discharge cycle. Time is
                                automatically calculated as tf=4000*(1/current), to ensure the
                                entire charge or discharge cycle is captured.
        p                   Defaults to none - present only for wrapping in an optimizer
        trim                Defaults to False, a value of true will remove the padded numbers
                                at the end, where False will allow padded values.

        Output              Description
        ------              ------------------------
        out                 A list of values for [time, voltage, current] of the simulation"""
        current = abs(current)
        if charge_first:
            solve = [
                self.charge(np.linspace(0, 5000 * (1 / current), n), p=p, trim=trim),
                self.discharge(np.linspace(0, 5000 * (1 / current), n), from_current_state=True, p=p, trim=trim)
            ]

            solve[-1].time += solve[0].time[-1]
            return np.concatenate(solve, axis=1)
        else:
            solve = [
                self.discharge(np.linspace(0, 5000 * (1 / current), n), p=p, trim=trim),
                self.charge(np.linspace(0, 5000 * (1 / current), n), from_current_state=True, p=p, trim=trim)
            ]

            solve[-1].time += solve[0].time[-1]
            return ChargeResult(np.concatenate([x.voltage for x in solve]),
                                np.concatenate([x.time for x in solve]),
                                np.concatenate([x.current for x in solve]),
                                None)

    def piecewise_current(
        self,
        times,
        currents,
        n_steps=50,
        from_current_state=False,
        p=None,
        internal=False,
    ):
        """This function wraps charge and discharge in order to chain them together to
        create the ability to simulate arbitrary piecewise currents. Only supports
        stair-style current stepping, ramps are not supported.
        Inputs
        -------     ----------------------
        times:      A list of values representing number of seconds at that current
        currents:   A list of values representing value of the current

        Example:

        spm.piecewise_current([50,100,30,40],[1,-0.5,.5,.3])

        where a negative current represents charging and a positive current
        represents discharge."""
        assert isinstance(times, list), "times must be a list"
        assert isinstance(currents, list), "Currents must be a list"
        assert len(times) == len(currents), "times and currents must be the same length"

        solve = []
        count = 0
        if not from_current_state:
            self.current_state = self.discharge_ICs
        for t, c in zip(times, currents):
            tt = np.linspace(0, t, n_steps)
            if c > 0:
                out = self.discharge(tt, current=c, from_current_state=True, p=p, internal=internal)
            else:
                out = self.charge(tt, current=c * -1, from_current_state=True, p=p, internal=internal)
            if count > 0:
                # add times together
                out = ChargeResult(out.time + solve[-1].time[-1], out.voltage, out.current, out.internal)
            solve.append(out)
            count += 1

        solve = ChargeResult(np.concatenate([x.time for x in solve]),
                             np.concatenate([x.voltage for x in solve]),
                             np.concatenate([x.current for x in solve]),
                             np.concatenate([x.internal for x in solve]) if internal else None)
        self.hist.append(solve)
        return solve

    def summary(self):
        """
        Returns a Latex render of the equations and node spacings
        used for the current model, in addition to a table
        containing the names and values of the parameters
        """

        return

    def opt_wrap(self, x0, currents_type, verbose=False):
        """The general optimization wrapping function - this function
        serves to call the numerical solution and calculate the
        root mean squared error at each experimental timepoint
        using cubic interpolation.  This ensures that the
        solver has sufficient accuracy while also allowing for
        direct comparisons with the experimental data."""
        from .fitting import rmse

        x = np.copy(self.initial)
        x[self.estimate_indices] = x0
        error = 0
        self.count += 1
        try:
            if currents_type == "constant":
                for t, v, c in zip(self.t_exp, self.v_exp, self.currents):
                    if c > 0:
                        self.current_state = self.discharge_ICs
                        solve = self.discharge(t, current=c, from_current_state=True, p=x)
                        error += rmse(solve.voltage, v)
                    else:
                        self.current_state = self.charge_ICs
                        solve = self.charge(t, current=-c, from_current_state=True, p=x)
                        error += rmse(solve.voltage, v)
            else:
                solve = self.piecewise_current(self.t_exp, self.currents, p=x)
                error += rmse(solve.voltage, self.v_exp)
            if verbose:
                print(error, x0)
        except IndexError:
            error = 100
            print(f"combination of parameters {x[self.estimate_indices]} failed")
        return error

    def fit(self, t_exp, v_exp, currents, currents_type="constant", method="Nelder-Mead", bounds=None, re=0, maxiter=100, tol=None):
        """Model-specific fitting function
        Parameters
        ----------
        t_exp: A numpy array containing the series of time values for experimental data

        v_exp: A numpy array containing the series of voltage values for experimental data

        method: A method to pass to Scipy.optimize.minimize, suggestions include:
            - 'Nelder-Mead'
            - 'SLSQP'
            - 'L-BFGS-B'

        bounds: Bounds for the given parameters
        """

        # enforce typing
        assert isinstance(t_exp, (np.ndarray, list)), "time array is not of type np.ndarray"
        assert isinstance(v_exp, (np.ndarray, list)), "voltage array is not of type np.ndarray"
        assert currents_type in [
            "constant",
            "piecewise",
        ], "currents_type should be either constant or piecewise"
        if currents_type == "constant":
            assert len(t_exp) == len(currents), "time and current lengths should match"
            assert len(t_exp) == len(v_exp), "time and voltage lengths should match"
        if currents_type == "piecewise":
            assert len(t_exp) == len(currents), "time samples and current samples should be of the same length"

        self.currents = currents
        self.currents_type = currents_type
        self.count = 0

        if re == 0:
            self.t_exp = t_exp
            self.v_exp = v_exp

        # # call the optimizer
        from scipy.optimize import minimize
        print(self.estimate_indices)
        print(self.initial[self.estimate_indices],)
        res = minimize(
            self.opt_wrap,
            self.initial[self.estimate_indices],
            args=(self.currents_type, self.verbose),
            bounds=bounds,
            method=method,
            options={"maxiter": int(maxiter)},
            tol=tol,
        )
        self.fitted_parameters = res.x
        self.initial_fit = True
        self.opt_result = res

        return self.opt_result

    def refit(self, method="Nelder-Mead", maxiter=100):
        assert self.initial_fit, "Please call fit before calling refit"
        self.initial[self.estimate_indices] = self.fitted_parameters
        self.fit(self.t_exp, self.v_exp, self.currents, method=method, re=1, maxiter=maxiter)
        return

    def generate_data(
        self,
        n,
        parameters,
        currents,
        loglist="auto",
        bounds=None,
        sample_type="sobol",
        distribution="uniform",
        sample_time=None,
        time_samples=100,
        summary=True,
        just_sample=False,
        verbose=False,
    ) -> SampleResult:
        """
        This function uses the existing Julia kernel to generate a set of data, similar to how the
        optimization function works. Since this julia kernel already exists, the calculations are note made in parallel.
        In the future, a separate file may exist which generates the data in parallel. It is recommended to call fit() first,
        in order to establish t_exp. Otherwise, time can be manually input.
        Parameters
        ----------
        n : the number of samples to make. For sobol, this is the total number, and for grid, this is the number of
                samples per grid dimension. Be careful using grid, as this can get out of hand very quickly (n^dim)

        loglist: A list of parameters which should be log spaced. Log spacing is advised for any variable changing
                    by more than 100x, i.e. max/min > 100.  If 'auto' is given, this will be detected automatically.
                    defaults to 'auto'.

        parameters: A list of parameters if desired modified parameters are different than the initialized set.

        bounds: Hard upper and lower limits for generating the data. If this is not given, +/- 30% is used. If a float is given,
                    +/- that amount is used. Custom bounds are recommended for good results, i.e., fewer failed samples.

        sample_type: defaults to 'sobol', options include 'grid' and 'saltelli'.

        distribution: defaults to 'uniform', describes the distribution of the sampling. Options include 'normal', 'uniform', and 'bi-modal'
                            (to be implemented)
        sample_time: The times to be interpolated for sampling. Filled values will be 2.5V, in keeping with the final
                        discharge voltage.  Defaults to linear spacing at 20% longer than t_exp
        time_samples: number of points in the linear timespacing, defaults to 100.
        """
        assert isinstance(n, int), "n must be an integer"
        assert isinstance(loglist, (str, list, bool)), 'loglist must be a list, False, or "auto"'
        assert sample_type in [
            "grid",
            "sobol",
            "saltelli",
            "random",
        ], "Available arguments for type are {}".format(["grid", "sobol", "saltelli", "random"])

        num_currents = len(currents)
        if num_currents > 1:
            print("multiple currents detected")
            # sort high to low in order to have failures (which are typically at high currents) block that data point
            currents = sorted(currents, reverse=True)

        # step 1 - handle the arguments
        generate_pars = parameters
        internal = False
        generate_inds = [i for i, x in enumerate(self.available_parameters) if x in generate_pars]
        assert generate_pars is not None
        print(generate_inds)

        # set up bounds (used for loglist)
        if bounds is None:
            self.bounds = [(x / 1.2, x * 1.2) for x in self.initial[generate_inds]]
        else:
            d1 = dict(zip(parameters, range(len(parameters))))
            bounds_inds = [d1[i] for i in [self.available_parameters[j] for j in generate_inds]]
            print(bounds_inds)
            self.bounds = [bounds[i] for i in bounds_inds]

        # set up log-spacing
        if loglist is False:
            self.loglist = list(np.zeros(len(generate_pars)))
        if isinstance(loglist, list):
            assert len(loglist) == len(
                generate_pars
            ), "expected loglist to be same length as generate_pars, \
            but got {} and {}".format(
                len(loglist), len(generate_pars)
            )
            self.loglist = loglist
        if loglist == "auto":
            self.loglist = [1 if x[0] / x[1] >= 100 else 0 for x in self.bounds]

        scaled_sample = self._generate_sample(n, sample_type, generate_pars)

        if just_sample:
            return SampleResult(scaled_sample, None, None, currents)

        voltages = []
        succeeded = []
        failed = []

        for parameter_set in scaled_sample:
            per_current_trial_voltage = []
            simulate_pars = np.copy(self.initial)
            simulate_pars[generate_inds] = self.log_descale_for_model(parameter_set)
            try:
                for i in currents:
                    if i > 0:
                        result = self.discharge(current=i, p=simulate_pars, internal=internal)
                        per_current_trial_voltage.append(result.voltage[:, np.newaxis])
                    else:
                        result = self.charge(current=-1 * i, p=simulate_pars, internal=internal)
                        per_current_trial_voltage.append(result.voltage[:, np.newaxis])
                succeeded.append(simulate_pars[generate_inds])
                voltages.append(np.concatenate(per_current_trial_voltage, axis=1))
            except IndexError:
                failed.append(simulate_pars[generate_inds])

        return SampleResult(np.array(succeeded), np.array(voltages), np.array(failed), currents)

    def _generate_sample(self, n, sample_type, generate_pars):
        # create the array using the spacing method of choice
        raw_sample = None
        if sample_type == "sobol":
            from sobol_seq import i4_sobol_generate

            raw_sample = i4_sobol_generate(len(generate_pars), n)
        elif sample_type == "saltelli":
            from SALib.sample import saltelli

            problem = {
                "names": generate_pars,
                "bounds": [[0, 1] for x in generate_pars],
                "num_vars": len(generate_pars),
            }
            raw_sample = saltelli.sample(problem, n, True)

        elif sample_type == "grid":
            from sklearn.utils.extmath import cartesian

            temp = np.linspace(0, 1, n)
            raw_sample = cartesian([temp for i in range(len(generate_pars))])

        elif sample_type == "random":
            raw_sample = np.random.random((n, len(generate_pars)))
        assert raw_sample is not None, "something went wrong - check that type is correct"
        print("expected shape is {}".format(raw_sample.shape))
        # map the raw array to bounds, adhering to log scaling rules
        scaled_sample = self.log_scale_matrix(raw_sample)
        return scaled_sample

    def abscale(self, matrix, a=-1, b=1):
        out = 0
        if matrix.shape == (matrix.shape[0],):
            matrix = matrix.reshape(-1, 1)
            out = 1
        new = np.zeros(matrix.shape)
        for i in range(matrix.shape[1]):
            new[:, i] = (b - a) * (matrix[:, i] - matrix[:, i].min()) / (matrix[:, i].max() - matrix[:, i].min()) + a
        if out == 0:
            return new
        else:
            return new[:, 0]

    def log_scale_matrix(self, matrix):
        """This function is used to resample the sampling methods into log
        space, if the value in loglist is a 1."""
        out = np.zeros(matrix.shape)
        for i in range(len(self.loglist)):
            if self.loglist[i] == 1:
                ub = np.log(self.bounds[i][1])
                lb = np.log(self.bounds[i][0])
                out[:, i] = self.abscale(matrix[:, i], ub, lb)
            else:
                ub = self.bounds[i][1]
                lb = self.bounds[i][0]
                out[:, i] = self.abscale(matrix[:, i], ub, lb)
        return out

    def log_descale_for_model(self, matrix):
        """This function takes in a matrix that has been log scaled and
        turns the log values back into normal values to pass to the model."""
        out = np.zeros(matrix.shape)
        for i in range(len(self.loglist)):
            if self.loglist[i] == 1:
                out[i] = np.exp(matrix[i])
            else:
                out[i] = matrix[i]
        return out

    def create_database(self, sample_result: SampleResult, filename: str, train_test_split: float = 0.3):
        """This function creates a dataset using pre-split data and saves it
        into a compressed h5 file with name filename. x, y, xt, yt are assumed
        to be numpy arrays."""
        import h5py

        with h5py.File(filename + ".hdf5", "w") as f:
            f.create_dataset("x", data=x, compression="gzip")
            f.create_dataset("xt", data=xt, compression="gzip")
            f.create_dataset("y", data=y, compression="gzip")
            f.create_dataset("yt", data=yt, compression="gzip")
        return

    def demonstrate_convergence(self, function):
        """This function seeks to demonstrate grid independence for the given
        charge / discharge pattern.  It will continue doubling the number of nodes
        in each region until the absolute error between subsequent discharges is less than 1e-6,
        giving an error if that does not occur.

        Parameters
        ----------
        function: A string describing the desired test

        Outputs
        -------
        Error as a function of nodes, as well as the minimum number of nodes needed
        for grid independence, or numerical convergence, at these conditions.

        """
        self.initial_discretization = self.initial[25:]
        self.nodes = []
        self.errors = []
        print(self.initial_discretization)
        for i in range(1, 10):
            self.nodes.append(self.initial_discretization * i)


import numpy as np
from .P2D_fd import model as P2D_fd


def p2d_fd(p, initial_state=(), tf=0, internal=False):
    model_inputs = np.concatenate([p, [0], [tf], initial_state])
    N1 = p[25]
    N2 = p[26]
    N3 = p[27]
    Nr1 = p[28]
    Nr2 = p[29]

    model_outputs = np.zeros((10000, int(5*N1+2*N2+5*N3+N1*Nr1+N3*Nr2+15)))
    P2D_fd(model_inputs, model_outputs)

    count = np.nonzero(model_outputs[:,-2])[0][-1]+1
    model_outputs = model_outputs[:count]
    final = model_outputs[-1]
    # need to select: time, voltage, current
    out = model_outputs[:,[0,-2,-1]]
    out[:,-1] /= 17.1
    # out[:,1] -= var[:,6]
    return out, final, model_outputs



import setuptools
import numpy

from distutils.command import build_ext
from setuptools import setup, Extension


USE_CYTHON = False

# ext = '.pyx'#  if USE_CYTHON else '.c'
ext = '.c'

ida_dir = "ampere/models/ida"
ida_files = ['ida.c', 'ida_band.c', 'ida_dense.c', 'ida_direct.c', 'ida_ic.c', 'ida_io.c', 'nvector_serial.c', 'sundials_band.c', 'sundials_dense.c', 'sundials_direct.c', 'sundials_math.c', 'sundials_nvector.c']
ida_requirements1 = [ida_dir + '/' + ida_file for ida_file in ida_files]


extensions = [
    Extension("ampere.models.P2D.P2D_fd", [f"ampere/models/P2D/P2D_fd{ext}", *ida_requirements1], include_dirs=['ampere/models/P2D/', numpy.get_include()]),
    Extension("ampere.models.SPM.SPM_fd", [f"ampere/models/SPM/SPM_fd{ext}", *ida_requirements1], include_dirs=[numpy.get_include()]),
    Extension("ampere.models.SPM.SPM_fd_sei", [f"ampere/models/SPM/SPM_fd_sei{ext}", *ida_requirements1], include_dirs=[numpy.get_include()]),
    Extension("ampere.models.SPM.SPM_par", [f"ampere/models/SPM/SPM_par{ext}", *ida_requirements1], include_dirs=[numpy.get_include()]),
]

# if USE_CYTHON:
# from Cython.Build import cythonize
# # # from Cython.Build import new_build_ext as build_ext
# extensions = cythonize(extensions, compiler_directives={'language_level': "3"})

with open("README.md", "r") as fh:
    long_description = fh.read()

# cmdclass = {'build_ext': build_ext}
# print()
print(setuptools.find_packages())
setup(
    name="ampere",
    version="0.5.9",
    author="Neal Dawson-Elli",
    author_email="nealde@uw.edu",
    description="A Python package for working with battery discharge data and physics-based battery models",

    # cmdclass=cmdclass,
    ext_modules=extensions,


    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nealde/Ampere",
    packages=[*setuptools.find_packages()],
    package_data={
        'ampere.models.P2D': ['ampere/models/P2D/*.c'],
        'ampere.models.SPM': ['ampere/models/SPM/*.c']
    },
    # include_package_data=True,
    install_requires=['cython', 'matplotlib < 3.4', 'numpy', 'scipy'],
    classifiers=[
        "Programming Language :: Python :: 3",
        'Programming Language :: Cython',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    keywords="battery numerical simulation modeling",
)



import os
import sys
import numpy as np
import h5py

sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))

from ampere import SingleParticleFDSEI, SingleParticleFD, SingleParticleParabolic, PseudoTwoDimFD


def test_values_unchanged():
    filename = 'ampere/tests/test_data'
    with h5py.File(filename + ".hdf5", "r") as f:
        for model in [SingleParticleParabolic(), PseudoTwoDimFD(), SingleParticleFD(), SingleParticleFDSEI()]:

            print(model.__class__.__name__)
            data = model.discharge(current=1.0)
            name = f'data{model.__class__.__name__}_discharge'
            new = np.concatenate([[data.time], [data.voltage], [data.current]], axis=0)
            # f.create_dataset(name, data=new, compression="gzip")
            expected = f.get(name)
            np.testing.assert_almost_equal(new, expected)

            data = model.charge(current=1.0)
            name = f'data{model.__class__.__name__}_charge'
            new = np.concatenate([[data.time], [data.voltage], [data.current]], axis=0)
            # f.create_dataset(name, data=new, compression="gzip")
            expected = f.get(name)
            np.testing.assert_almost_equal(new, expected)

            data = model.discharge(current=1.0, internal=True)
            if data.internal is not None:
                name = f'data{model.__class__.__name__}_discharge_internal'
                # f.create_dataset(name, data=data.internal, compression="gzip")
                expected = f.get(name)
                np.testing.assert_almost_equal(data.internal, expected)

            data = model.piecewise_current([500,30,60,100], [1.0,-0.2,0.5,-2])
            name = f'data{model.__class__.__name__}_piecewise'
            new = np.concatenate([[data.time], [data.voltage], [data.current]], axis=0)
            # f.create_dataset(name, data=new, compression="gzip")
            expected = f.get(name)
            np.testing.assert_almost_equal(new, expected)


def test_optimization():
    labels = ['Rp', 'lp', 'kp']
    values = [2.0e-6, 78e-6, 2.334e-9]
    p1 = dict(zip(labels, values))

    spm1 = SingleParticleFDSEI()
    spm2 = SingleParticleFDSEI(initial_parameters=p1, estimate_parameters=['Rp', 'lp', 'kp'])

    # create the experimental data
    a = [
        spm1.discharge(current=0.5, trim=True),
        spm1.charge(current=1.0, trim=True)
    ]

    t_exp = [x.time for x in a]
    v_exp = [x.voltage for x in a]
    res = spm2.fit(t_exp, v_exp, currents=[0.5, -1.0], maxiter=1000)

    np.testing.assert_almost_equal(res.x, [7.84309206e-05, 2.01704867e-06, 2.36852421e-09])


def test_generate_data():
    spm = SingleParticleFDSEI()
    data = spm.generate_data(100, currents=[0.5, 1.0], parameters=['Rp', 'Rn'], bounds=[[1e-12, 1e-3], [1e-12, 1e-3]], sample_type='random')
    assert data.inputs.shape == (98, 2)
    assert data.voltages.shape == (98, 500, 2)
    assert data.failed.shape == (2, 2)


test_values_unchanged()
test_optimization()
test_generate_data()


import numpy as np


def rmse(a, b):
    """
    A function which calculates the root mean squared error
    between two vectors.

    Notes
    ---------
    .. math::

        RMSE = \\sqrt{\\frac{1}{n}(a-b)^2}
    """
    assert isinstance(a, np.ndarray), \
        'This function expects a numpy array as input'
    assert isinstance(b, np.ndarray), \
        'This function expects a numpy array as input'
    assert len(a) == len(b), \
        'a and b should be the same length, but got %i, %i' % (len(a), len(b))
    return np.sqrt(np.mean(np.square(a-b)))

